
# -- import packages: --------------------------------------------------------------------
from typing import Union, Dict, List
import inspect


# -- operational class: ------------------------------------------------------------------
class KwargExtractor:
    def __init__(self, func):
        self.func = func

    # -- methods: ------------------------------------------------------------------------
    def __parse__(self, kwargs, parse_ignore=["self"]):
        self._PASSED = {}

        if isinstance(kwargs, dict):
            for key, val in kwargs.items():
                if not key in self._ignore:
                    self._PASSED[key] = val
        elif isinstance(kwargs, list):
            for key in kwargs:
                self._PASSED[key] = None

    def _extract_func_params(self):
        return list(inspect.signature(self.func).parameters.keys())

    def query(self):
        self.func_kwargs = {}

        for key, val in self._PASSED.items():
            if not key in self._ignore:
                if key in self.func_params:
                    if self._obj:
                        val = getattr(self._obj, key)
                    #                     print(key, "...regular")
                    self.func_kwargs[key] = val

            if key == "kwargs":
                if not val is None:
                    self._check_literal_kwargs(val)

        return self.func_kwargs

    def _check_literal_kwargs(self, kwargs):
        for key, val in kwargs.items():
            if not key in self._ignore:
                #                 print(key, "...literal kwargs")
                self.func_kwargs[key] = val

    # -- property: -----------------------------------------------------------------------
    @property
    def func_params(self):
        return self._extract_func_params()

    def __call__(self, kwargs: Union[Dict, List], obj=None, ignore=["self", "kwargs"]):
        self._obj = obj

        if (
            (not self._obj is None)
            and (not isinstance(kwargs, dict))
            and (not isinstance(kwargs, list))
        ):
            kwargs = self._obj.__dir__()

        self._ignore = ignore
        self.__parse__(kwargs, parse_ignore=["self"])

        return self.query()


# -- API-facing function: ----------------------------------------------------------------
def function_kwargs(func, kwargs=None, obj=None, ignore=["self"]):
    """
    Returns the subset of kwargs that can be used in the func.

    Parameters:
    -----------
    func

    kwargs
        if obj is passed, this argument is overridden.

    obj
        if kwargs is passed, obj overrides.

    Returns:
    --------
    function_kwargs
        type: list
    """

    kwarg_extractor = KwargExtractor(func)
    return kwarg_extractor(kwargs, obj)




__module_name__ = "_abc_parse.py"
__doc__ = """Better abstract base class for auto-parsing inputs."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["mvinyard.ai@gmail.com"])


# -- import packages: ----------------------------------------------------------
import abc


# -- set typing: ---------------------------------------------------------------
from typing import Dict, List, Any, Tuple


# -- Controller class: ---------------------------------------------------------
class ABCParse(abc.ABC):
    _BUILT = False

    def __init__(self, *args, **kwargs):
        """
        we avoid defining things in __init__ because this subsequently
        mandates the use of `super().__init__()`
        
        Example
        -------
        ```
        class DataConfiguration(utils.ABCParse):
            def __init__(self, x=2, y=3, *args, **kwargs):
                self.__parse__(locals(), public=[None])

            def __call__(self, x=4, y=5, z=3, *args, **kwargs):
                self.__update__(locals(), private=[None])
        
        
        dc = DataConfiguration(alpha=0.2)
        dc._PARAMS
        dc(beta=0.4)
        dc._PARAMS
        dc._kwargs
        dc._PARAMS
        ```
        """
        pass

    def __build__(self) -> None:
        self._PARAMS = {}
        self._IGNORE = ["self", "__class__"]
        self._stored_private = []
        self._stored_public = []

        self._BUILT = True

    def __set__(
        self, key: str, val: Any, public: List = [], private: List = []
    ) -> None:
        self._PARAMS[key] = val
        
        if (key in private) and (not key in public):
            self._stored_private.append(key)
            key = f"_{key}"
        else:
            self._stored_public.append(key)
        setattr(self, key, val)

    def __set_existing__(self, key: str, val: Any) -> None:
        
        passed_key = key

        if key in self._stored_private:
            key = f"_{key}"

        if passed_key == "kwargs":
            attr = getattr(self, key)
            attr.update(val)
            setattr(self, key, attr)
            self._PARAMS.update(val)
            
        elif passed_key == "args":
            attr = getattr(self, key)
            attr += val
            setattr(self, key, attr)
            self._PARAMS[passed_key] += val
            
        else:
            self._PARAMS[passed_key] = val
            setattr(self, key, val)

    @property
    def _STORED(self) -> List:
        return self._stored_private + self._stored_public

    def __setup_inputs__(self, kwargs, public, private, ignore) -> Tuple[List]:
        if not self._BUILT:
            self.__build__()

        self._IGNORE += ignore

        if len(public) > 0:
            private = list(kwargs.keys())

        return public, private

    def __parse__(
        self, kwargs: Dict, public: List = [], private: List = [], ignore: List = []
    ) -> None:
        """Central function of this class"""

        public, private = self.__setup_inputs__(kwargs, public, private, ignore)

        for key, val in kwargs.items():
            if not key in self._IGNORE:
                self.__set__(key, val, public, private)

    def __update__(
        self, kwargs: Dict, public: List = [], private: List = [], ignore: List = []
    ) -> None:
        """To be called after __parse__ has already been called."""

        public, private = self.__setup_inputs__(kwargs, public, private, ignore)

        for key, val in kwargs.items():
            if not (val is None) and (key in self._STORED):
                self.__set_existing__(key, val)

            elif not (val is None) and not (key in self._IGNORE):
                self.__set__(key, val, public, private)




import setuptools
import re
import os
import sys


setuptools.setup(
    name="ABCParse",
    version="0.0.6rc0",
    python_requires=">3.9.0",
    author="Michael E. Vinyard",
    author_email="mvinyard@broadinstitute.org",
    url=None,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    description="A better base class to automatically parse local args.",
    packages=setuptools.find_packages(),
    install_requires=[],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="MIT",
)
